public inbox for gcc-cvs@sourceware.org
help / color / mirror / Atom feed
* [gcc r14-9819] Copy condition->expr map when inlining [PR114599]
@ 2024-04-06 11:11 J?rgen Kvalsvik
  0 siblings, 0 replies; only message in thread
From: J?rgen Kvalsvik @ 2024-04-06 11:11 UTC (permalink / raw)
  To: gcc-cvs

https://gcc.gnu.org/g:c6892a430a9752aea167265621c3ae7a3e11159f

commit r14-9819-gc6892a430a9752aea167265621c3ae7a3e11159f
Author: Jørgen Kvalsvik <j@lambda.is>
Date:   Fri Apr 5 21:42:07 2024 +0200

    Copy condition->expr map when inlining [PR114599]
    
    When a function is tree-inlined, copy the condition -> expression mapping
    from the inlined function into the caller, shifted so uids are not
    mixed.  Tree inlining was always problematic under condition coverage -
    either through a nullptr dereference (triggered by the test case), or
    through quietly mixing caller conditions with the callee conditions.
    
            PR middle-end/114599
    
    gcc/ChangeLog:
    
            * tree-inline.cc (add_local_variables): Copy cond_uids mappings.
    
    gcc/testsuite/ChangeLog:
    
            * gcc.misc-tests/gcov-pr114599.c: New test.

Diff:
---
 gcc/testsuite/gcc.misc-tests/gcov-pr114599.c | 25 +++++++++++++++++++++++++
 gcc/tree-inline.cc                           | 20 +++++++++++++++++++-
 2 files changed, 44 insertions(+), 1 deletion(-)

diff --git a/gcc/testsuite/gcc.misc-tests/gcov-pr114599.c b/gcc/testsuite/gcc.misc-tests/gcov-pr114599.c
new file mode 100644
index 00000000000..e4c78c9c121
--- /dev/null
+++ b/gcc/testsuite/gcc.misc-tests/gcov-pr114599.c
@@ -0,0 +1,25 @@
+/* PR middle-end/114599 */
+/* { dg-do compile } */
+/* { dg-options "-fcondition-coverage" } */
+
+/* Check that a function with a condition inlined into a function without a
+   conditional works.  When inlining happens the condition -> expression
+   mapping must be carried over.  */
+
+extern int type;
+
+void fn (void);
+
+__attribute__((always_inline))
+inline void
+do_all_fn_doall_arg (void)
+{
+  if (type)
+    fn ();
+}
+
+void
+do_all_fn_LHASH_DOALL_ARG_arg2 (void)
+{
+  do_all_fn_doall_arg ();
+}
diff --git a/gcc/tree-inline.cc b/gcc/tree-inline.cc
index eebcea8a029..b18917707cc 100644
--- a/gcc/tree-inline.cc
+++ b/gcc/tree-inline.cc
@@ -4659,7 +4659,8 @@ prepend_lexical_block (tree current_block, tree new_block)
   BLOCK_SUPERCONTEXT (new_block) = current_block;
 }
 
-/* Add local variables from CALLEE to CALLER.  */
+/* Add local variables from CALLEE to CALLER.  If set for condition coverage,
+   copy basic condition -> expression mapping to CALLER.  */
 
 static inline void
 add_local_variables (struct function *callee, struct function *caller,
@@ -4689,6 +4690,23 @@ add_local_variables (struct function *callee, struct function *caller,
 	  }
 	add_local_decl (caller, new_var);
       }
+
+  /* If -fcondition-coverage is used and the caller has conditions, copy the
+     mapping into the caller but and the end so the caller and callee
+     expressions aren't mixed.  */
+  if (callee->cond_uids)
+    {
+      if (!caller->cond_uids)
+	caller->cond_uids = new hash_map <gcond*, unsigned> ();
+
+      unsigned dst_max_uid = 0;
+      for (auto itr : *callee->cond_uids)
+	if (itr.second >= dst_max_uid)
+	  dst_max_uid = itr.second + 1;
+
+      for (auto itr : *callee->cond_uids)
+	caller->cond_uids->put (itr.first, itr.second + dst_max_uid);
+    }
 }
 
 /* Add to BINDINGS a debug stmt resetting SRCVAR if inlining might

^ permalink raw reply	[flat|nested] only message in thread

only message in thread, other threads:[~2024-04-06 11:11 UTC | newest]

Thread overview: (only message) (download: mbox.gz / follow: Atom feed)
-- links below jump to the message on this page --
2024-04-06 11:11 [gcc r14-9819] Copy condition->expr map when inlining [PR114599] J?rgen Kvalsvik

This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox;
as well as URLs for read-only IMAP folder(s) and NNTP newsgroup(s).